module nvec-map where

-- N-ary MAP just for vectors

open import Data.Vec hiding (_⊛_)
open import Data.Vec1
open import Data.List hiding (zip)
open import Data.Nat
open import Data.Unit
open import Data.Product hiding (zip)
open import Data.Bool

open import local-lib

arrTy : {n : ℕ} → Vec₁ Set (suc n) → Set
arrTy {0}      (A ∷ [])  = A
arrTy {suc n}  (A ∷ As)  = A → arrTy As

arrTyVec  : {m n : ℕ} → Vec₁ Set (suc n) → Set
arrTyVec {m} As = 
  arrTy (repeat₁ (\ A → Vec A m) ⊛₁ As )

nvec-map  : {m : ℕ} → (n : ℕ) 
          → (As : Vec₁ Set (suc n)) 
          → arrTy As → arrTyVec {m} As

nvec-map n As f = g {n} As (repeat f) where
  g  : {n m : ℕ} 
     → (As : Vec₁ Set (suc n)) 
     → Vec (arrTy As) m → arrTyVec {m} As
  g  {0}      (A ∷ [])   a  = a
  g  {suc n}  (A1 ∷ As)  f  = 
        (\ a → g As (f ⊛ a))

nmap  : (n : ℕ) → {m : ℕ}
      → ∀⇒₀ (\(As : Vec₁ Set (suc n))
            → arrTy As → arrTyVec {m} As)
nmap n {m} =  \⇒₀ (nvec-map {m} n)

vec-map1 : {A B : Set} {n : ℕ} → (A → B) → (Vec A n → Vec B n)
vec-map1 = ↓ (nmap 1)

vec-map2 : {A B C : Set} {n : ℕ} → (A → B → C) 
         → (Vec A n → Vec B n → Vec C n)
vec-map2 = ↓ (nmap 2)


vt1 = vec-map1 (\ x → x + 1) (1 ∷ 2 ∷ 3 ∷ [])
vt2 = vec-map2 (\ x y → x + y) (1 ∷ 2 ∷ 3 ∷ []) (1 ∷ 2 ∷ 3 ∷ [])
