module towards-genericity where

open import Data.Unit
open import Data.Nat
open import Data.Vec hiding (_⊛_)
open import Data.Vec1
open import Data.Product
open import Data.Sum
open import local-lib
open import nvec-map

data Tyc : Set where
  Nat   : Tyc
  Unit  : Tyc
  Prod  : Tyc → Tyc → Tyc
  Arr   : ℕ → Tyc → Tyc
  Var   : Tyc

⌊_⌋                : Tyc → (Set → Set)
⌊ Nat ⌋         a  = ℕ
⌊ Unit ⌋        a  = ⊤
⌊ Prod t1 t2 ⌋  a  = ⌊ t1 ⌋ a × ⌊ t2 ⌋ a
⌊ Arr n t1 ⌋    a  = Vec (⌊ t1 ⌋ a) n
⌊ Var ⌋         a  = a

grepeat : (t : Tyc) → { a : Set } → a → ⌊ t ⌋ a
grepeat Nat  a   = 0 
grepeat Unit a   = tt
grepeat (Prod t1 t2) a = ( grepeat t1 a , grepeat t2 a )
grepeat (Arr n t1) a = repeat {n} (grepeat t1 a)
grepeat Var a = a

gzap  : (t : Tyc) → { a b : Set } → ⌊ t ⌋ (a → b) 
      → ⌊ t ⌋ a → ⌊ t ⌋ b
gzap Nat _ _ = 0
gzap Unit _ _ = tt
gzap (Prod t1 t2) (f1 , f2) (x1 , x2) = (gzap t1 f1 x1 , gzap t2 f2 x2)
gzap (Arr n t1) v1 v2 = repeat (gzap t1) ⊛ v1 ⊛ v2 
gzap Var f x = f x

gvec-map  : (t : Tyc) → (n : ℕ) 
          → (As : Vec₁ Set (suc n)) 
          → arrTy As → arrTy (repeat₁ ⌊ t ⌋ ⊛₁ As)
gvec-map t n As f = g As (grepeat t f) where
  g  : {n : ℕ} → (As : Vec₁ Set (suc n)) 
     → ⌊ t ⌋ (arrTy As) 
     → arrTy (repeat₁ ⌊ t ⌋ ⊛₁ As)
  g  {0}      (A ∷ [])   a   = a
  g  {suc n}  (A1 ∷ As)  f   = 
        (\ a → g As (gzap t f a))

gmap : (t : Tyc) → (n : ℕ) 
     → ∀⇒₀ (\(As : Vec₁ Set (suc n))
           → arrTy As → arrTy (repeat₁ ⌊ t ⌋ ⊛₁ As))
gmap t n = \⇒₀ (gvec-map t n)

example-map  : {m : ℕ} → { A B : Set } → (A → B) 
             → Vec (A × (A × ⊤)) m
             → Vec (B × (B × ⊤)) m
example-map = 
  ↓ (gmap (Arr _ (Prod Var (Prod Var Unit))) 1 )

