module local-lib where

-- generic library functions

open import Data.Unit
open import Data.Product
open import Data.Nat
open import Data.Sum
open import Data.Vec1
open import Data.Bool
open import Data.Vec hiding (_⊛_)

import Relation.Binary.PropositionalEquality as PropEq
import Relation.Binary.PropositionalEquality1 as PropEq1
open PropEq using (_≡_)
open PropEq1 using (_≡₁_)

-- Array type
Array : ℕ → Set → Set
Array n A = Vec A n

infixl 50 _⊛₁_
infixl 50 _⊛_

repeat : {n : ℕ} → {A : Set} → A → Vec A n
repeat {zero}    x = []
repeat {(suc n)} x = x ∷ repeat {n} x

_⊛_ : {A B : Set} {n : ℕ} → Vec (A → B) n → Vec A n → Vec B n
[] ⊛ [] = [] 
(a ∷ As) ⊛ (b ∷ Bs) = (a b ∷ As ⊛ Bs)

repeat₁ : {n : ℕ} → {A : Set1} → A → Vec₁ A n
repeat₁ {zero}    x = []
repeat₁ {(suc n)} x = x ∷ repeat₁ {n} x

_⊛₁_ : {a b : Set1}{n : ℕ} → Vec₁ (a → b) n → Vec₁ a n → Vec₁ b n
[] ⊛₁ [] = [] 
(a ∷ as) ⊛₁ (b ∷ bs) = (a b ∷ as ⊛₁ bs)
 
≡-tail₁  : ∀ {A} {n} {t1 t2 : Vec₁ A n} {x : A} 
        → t1 ≡₁ t2 
        → _≡₁_ {Vec₁ A (suc n)} (x ∷ t1)  (x ∷ t2)
≡-tail₁ PropEq1.refl = PropEq1.refl

≡-tail  : ∀ {A} {n} {t1 t2 : Vec A n} {x : A} 
        → t1 ≡ t2 
        → _≡_ {Vec A (suc n)} (x ∷ t1) (x ∷ t2)
≡-tail PropEq.refl = PropEq.refl

≡-pair :  {A : Set}{B : Set}{ x x' : A}{y y' : B} 
  → x ≡ x' → y ≡ y' → _,_ {A}{\ x → B} x y ≡ (x' , y')
≡-pair PropEq.refl PropEq.refl = PropEq.refl


eqNat : ℕ → ℕ → Bool
eqNat zero zero = true
eqNat (suc n) (suc m) = eqNat n m
eqNat _ _ = false


-- Tuples. Data structures indexed by a vector of types

prodTy : {n : ℕ} → (As : Vec₁ Set n) → Set
prodTy {0}            _         = ⊤
prodTy {1}            (A ∷ [])  = A
prodTy {suc (suc _)}  (A ∷ As)  = (A × prodTy As)

tuple-repeat : {n : ℕ}{A : Set} → A → prodTy {suc n} (repeat₁ A)
tuple-repeat {zero}  a = a
tuple-repeat {suc n} a = (a , tuple-repeat {n} a) 

tuple-⊛ : {n : ℕ}{As Bs : Vec₁ Set (suc n) } → 
  prodTy (repeat₁ {suc n} (\ A B → A → B) ⊛₁ As ⊛₁ Bs) → 
  prodTy As → prodTy Bs 
tuple-⊛ {zero} {A ∷ []}{B ∷ []} f a  =  f a
tuple-⊛ {suc n}{A ∷ As}{B ∷ Bs} (f , fs) (a , as) = ( f a , tuple-⊛ {n} fs as ) 

tuple-unzip : {n : ℕ} {As Bs : Vec₁ Set (suc n)} → 
              prodTy ((repeat₁ _×_ ⊛₁ As) ⊛₁ Bs) → 
              prodTy As × prodTy Bs
tuple-unzip {zero}{A ∷ []}{B ∷ []} (a , b) = (a , b)
tuple-unzip {suc n}{(A ∷ As)}{(B ∷ Bs)} ( (a , b) , abs) with tuple-unzip {n} abs
... | ( as , bs ) = ( (a , as) , (b , bs) )

tuple-zip : {n : ℕ} {As Bs : Vec₁ Set (suc n)} → 
              prodTy As → prodTy Bs →
              prodTy ((repeat₁ _×_ ⊛₁ As) ⊛₁ Bs) 
tuple-zip {zero}{(A ∷ [])}{(B ∷ [])} a b = (a , b)
tuple-zip {suc n}{(A ∷ As)}{(B ∷ Bs)}   (a , as)  (b , bs) = 
  ( (a , b) , tuple-zip {n} as bs )


tuple-arr-transpose : {m : ℕ} → {n : ℕ} → {As : Vec₁ Set (suc n)}
  → Array m (prodTy As) 
  → prodTy (repeat₁ (Array m) ⊛₁ As)
tuple-arr-transpose {zero}{n}{As}  [] = k As where
              k : {n : ℕ} → (As : Vec₁ Set (suc n)) → prodTy (repeat₁ (Array 0) ⊛₁ As)
              k {zero} (A ∷ []) = []
              k {suc n} (A ∷ As) = ( [] , k As )
tuple-arr-transpose {suc m} {n}{As} (x ∷ xs) = 
    k As x (tuple-arr-transpose {m}{n}{As} xs) where
              k : {n : ℕ} (As : Vec₁ Set (suc n)) → (prodTy As) → 
                         prodTy (repeat₁ (Array m) ⊛₁ As) → 
                         prodTy (repeat₁ (Array (suc m)) ⊛₁ As)
              k {zero} (A ∷ []) x1  v =  x1 ∷ v
              k {suc n} (A ∷ As) (x , xs) (v , vs) =
                    x ∷ v , k As xs vs

-- tuples of arrays to arrays of tuples. 
-- note the Vec.map2 in the last line.
arr-tuple-transpose : {m : ℕ} → {n : ℕ} → {As : Vec₁ Set (suc n)}
  → prodTy (repeat₁ (Array m) ⊛₁ As)
  → Array m (prodTy As) 
arr-tuple-transpose {m}{zero}{A ∷ []} x = x 
arr-tuple-transpose {m}{suc n}{A ∷ As} (x , xs) =
   ((replicate _,_) ⊛ x) ⊛ arr-tuple-transpose {m}{n}{As} xs




-- Dependent vectors. Isomorphic to tuples but with inductive types
-- instead of fixpoints. Trying to figure out which is better to 
-- support. Note that this has to live in Set1 instead of Set, while 
-- the tuples don't.

data DVec : {n : ℕ} → (As : Vec₁ Set n) → Set1 where 
  nil  : DVec []
  _::_ : {n : ℕ}{A : Set}{As : Vec₁ Set n} → A → DVec As → DVec (A ∷ As)     --- note not unicode symbol

-- would be nice if we could automatically generate these :-)
dvec-repeat : {n : ℕ}{A : Set} → A → DVec {n} (repeat₁ A)
dvec-repeat {zero}  a = nil
dvec-repeat {suc n} a = (a :: dvec-repeat {n} a) 

{-
-- Need ×₁ in result, otherwise everything works.
dvec-unzip : {n : ℕ} {As Bs : Vec₁ Set n} → 
              DVec ((repeat₁ _×_ ⊛₁ As) ⊛₁ Bs) → 
              DVec As ×₁ DVec Bs
dvec-unzip {zero}{[]}{[]} _ = (nil , nil)
dvec-unzip {suc n}{(A ∷ As)}{(B ∷ Bs)} ( (a , b) :: abs) with dvec-unzip {n} abs
... | ( as , bs ) = ( (a :: as) , (b :: bs) )
-}
dvec-zip : {n : ℕ} {As Bs : Vec₁ Set n} → 
              DVec As → DVec Bs →
              DVec ((repeat₁ _×_ ⊛₁ As) ⊛₁ Bs) 
dvec-zip {zero}{[]}{[]} _ _ = nil
dvec-zip {suc n}{(A ∷ As)}{(B ∷ Bs)}   (a :: as)  (b :: bs) = 
  ( (a , b) :: dvec-zip {n} as bs )


-- Transforming a vector type As → B into A1 → A2 → ... → B
-- not used for now, but it is nice notation that we might want 
-- to adopt elsewhere. Unfortunately, my LaTeX installation cannot 
-- print that double arrow. 

_↠_ : {n : ℕ} → (As : Vec₁ Set n) → Set1 → Set1
[] ↠ B = B
(A ∷ As) ↠ B = A → (As ↠ B)

curryTuple : {n : ℕ} {As : Vec₁ Set (suc n)} {B : Set1} → 
   ((prodTy As) → B) → (As ↠ B ) 
curryTuple {zero} {(A ∷ [])} x = x
curryTuple {suc m} {(A ∷ As)} f = 
    \ x → curryTuple {m}{As} (\ xs → f ( x , xs))

curryDVec : {n : ℕ} {As : Vec₁ Set n} {B : Set1} → 
   ((DVec As) → B) → (As ↠ B ) 
curryDVec {zero} {[]} x = (x nil)
curryDVec {suc m} {(A ∷ As)} f = 
    \ x → curryDVec {m}{As} (\ xs → f ( x :: xs))

-- quantify
∀⇒ : {n : ℕ} {A : Set1} → (Vec₁ A n → Set1) → Set1
∀⇒ {zero}       B = B []
∀⇒ {suc n} {A}  B = {a : A} → ∀⇒ {n} (\ as → B (a ∷ as))
  
--- curry
\⇒  : {n : ℕ} {A : Set1} {B : Vec₁ A n → Set1} 
    → ((X : Vec₁ A n) → B X) → (∀⇒ B)
\⇒ {zero}       f = f []
\⇒ {suc n} {A}  f = \ {a : A} → \⇒ {n} (\ as → f (a ∷ as)) 

-- uncurry
/⇒ : (n : ℕ) → {K : Set1}{B : Vec₁ K n → Set1} 
  → (∀⇒ B) → (A : Vec₁ K n) → B A 
/⇒ (zero) f  []       = f
/⇒ (suc n) f (a ∷ as) = /⇒ n (f {a}) as

data ⇑ : Set → Set1 where
  ↑ : { A : Set } → A → ⇑ A 

↓ : { A : Set } → ⇑ A  → A 
↓ (↑ x) = x

⟰ : {A : Set1} → (A → Set) → (A → Set1)
⟰ f x = ⇑ (f x)
