{-# OPTIONS --type-in-type #-}

-- Doubly-generic version of unzipWith
module ngsplit where

open import Data.Unit
open import Data.Product hiding (map)
open import Data.Nat 
open import Data.Sum hiding (map)
open import Data.List hiding ([_]; replicate; map)
open import Data.Vec hiding ([_]; _⊛_)
open import Data.Vec1 hiding (map)
open import Data.Bool
open import Data.Unit1
open import Data.Maybe hiding (maybe)

open import local-lib
open import data-view
open import nvec-map

NGsplit : {n : ℕ} → (v : Vec₁ Set (suc n)) → Set1
NGsplit (A1 ∷ As) = A1 → prodTy As

duplicate  : {t : Set} → (n : ℕ) → t 
           → prodTy {suc n} (repeat₁ t)
duplicate 0        x = x 
duplicate (suc n)  x = (x , duplicate n x) 

defNat : (n : ℕ) → NGsplit {suc n} (repeat₁ ℕ)
defNat n = \ x → duplicate n x

defUnit : (n : ℕ) → NGsplit {suc n} (repeat₁ ⊤)
defUnit n = \ x → duplicate n x

prodn  : {n : ℕ} → (As Bs : Vec₁ Set n)
       → prodTy As → prodTy Bs 
       → prodTy (repeat₁ _×_ ⊛₁ As ⊛₁ Bs)
prodn {0}  _        _         a b  = tt
prodn {1} (A ∷ [])  (B ∷ [])  a b  = (a , b)
prodn {suc (suc n)}  (A ∷ As)  (B ∷ Bs)  (a , as)  (b , bs)  =
  ( (a , b) , prodn {suc n} _ _ as bs )     

defPair  : (n : ℕ) 
         → (As : Vec₁ Set (suc n)) → (NGsplit As) 
         → (Bs : Vec₁ Set (suc n)) → (NGsplit Bs) 
         → NGsplit (repeat₁ _×_ ⊛₁ As ⊛₁ Bs)
defPair n (A ∷ As) a (B ∷ Bs) b =
  \ p → prodn {n} _ _ (a (proj₁ p)) (b (proj₂ p)) 

injLeft  : {n : ℕ} {As Bs : Vec₁ Set (suc n)} 
         → prodTy As
         → prodTy (repeat₁ _⊎_ ⊛₁ As ⊛₁ Bs)
injLeft {0}      {A ∷ []}  {B ∷ []}  a         = inj₁ a
injLeft {suc n}  {A ∷ As}  {B ∷ Bs}  (a , as)  = 
  ( inj₁ a , injLeft {n} as )

injRight  : {n : ℕ} {As Bs : Vec₁ Set (suc n)} 
          → prodTy Bs
          → prodTy (repeat₁ _⊎_ ⊛₁ As ⊛₁ Bs)
injRight {0}      {A ∷ []}  {B ∷ []}  b         = inj₂ b
injRight {suc n}  {A ∷ As}  {B ∷ Bs}  (b , bs)  = 
  ( inj₂ b , injRight {n} bs )

defSum  : (n : ℕ) 
        → (As : Vec₁ Set (suc n)) → (NGsplit As)
        → (Bs : Vec₁ Set (suc n)) → (NGsplit Bs)
        → NGsplit (repeat₁ _⊎_ ⊛₁ As ⊛₁ Bs)
defSum 0        (A ∷ [])  af  (B ∷ [])  bf = \ _ → tt
defSum (suc n)  (A ∷ As)  af  (B ∷ Bs)  bf = f
  where  f : A ⊎ B → prodTy (repeat₁ _⊎_ ⊛₁ As ⊛₁ Bs)
         f (inj₁ x1)  = injLeft {n} (af x1) 
         f (inj₂ x1)  = injRight {n} (bf x1)

nwayOne : (k : Kind) → {A : ⟦ k ⟧} 
  → NGsplit ⟨ k ⟩ (A ∷ [])
nwayOne ⋆ = \x → record {}
nwayOne (k1 ⇒ k2) {A} = \ {a} kita → h (a ∷ [])
  where h : (a : Vec₁ (⟦ k1 ⟧) 1) 
             → NGsplit ⟨ k2 ⟩ ( (A ∷ []) ⊛₁ a )
        h ( B ∷ [] ) = nwayOne k2

split-const : {n : ℕ} → TyConstEnv {n} NGsplit
split-const {zero}{k} _  = nwayOne k
split-const {suc n} Nat  = defNat n
split-const {suc n} Unit = defUnit n 
split-const {suc n} Prod = k-curry (⋆ ⇒ ⋆ ⇒ ⋆) (defPair (suc n))
split-const {suc n} Sum  = k-curry (⋆ ⇒ ⋆ ⇒ ⋆) (defSum (suc n))

prodTy-data : {n : ℕ} → DataGen {n} prodTy
prodTy-data {0}      dt  (e ∷ [])  bt           = DT-to dt bt
prodTy-data {suc n}  dt  (e ∷ eq)  (bt1 , bts)  = 
  DT-to dt bt1 , prodTy-data dt eq bts

split-data : {n : ℕ} → DataGen {n} NGsplit
split-data {0}      dt  (e ∷ [])  bt = 
  \ s → bt (DT-from dt s)
split-data {suc n}  dt  (e ∷ eq)  bt = 
  \ s → prodTy-data dt eq (bt (DT-from dt s))

ngsplit  : (n : ℕ) → {k : Kind} → (e : Ty k)
         → NGsplit ⟨ k ⟩ (repeat₁ {suc n} ⌊ e ⌋)
ngsplit n e = ngen e split-const split-data

unzipWithMaybe3 : { A B C : Set } → (A → B × C) → 
  ( Maybe A → Maybe B × Maybe C )
unzipWithMaybe3 = ngsplit 2 maybe

unzipWithMaybe4 : { A B C D : Set } → (A → B × C × D) → 
  ( Maybe A → Maybe B × Maybe C × Maybe D)
unzipWithMaybe4 = ngsplit 3 maybe 

unzipMaybe2 : { A B : Set } → Maybe (A × B) →
   (Maybe A × Maybe B)
unzipMaybe2 = unzipWithMaybe3 (\ x → x)

unzipMaybe3 : { A B C : Set } → Maybe (A × B × C) →
   (Maybe A × Maybe B × Maybe C)
unzipMaybe3 = unzipWithMaybe4 (\ x → x)

unzipWith2  : {A B C : Set} {m : ℕ} → (A → B × C)
            → Vec A m → Vec B m × Vec C m
unzipWith2 = ngsplit 2 vec

unzip2  : {A B C : Set} {m : ℕ} 
        → Vec (B × C) m → Vec B m × Vec C m
unzip2 = unzipWith2 (\ x → x)

unzipWith3  : {A B C D : Set} {m : ℕ} 
            → (A → B × C × D) → Vec A m 
            → Vec B m × Vec C m × Vec D m
unzipWith3 = ngsplit 3 vec

-- The following code shows how complicated it is to define split
-- directly for vectors.

-- Given a vector of types, create a product 
-- where every element of the product is an empty 
-- vector of the corresponding type in the array.
prodNil  : {n : ℕ} → (As : Vec₁ Set n) 
         → prodTy (repeat₁ (Array 0) ⊛₁ As)
prodNil {0}            []        = tt
prodNil {1}            (A ∷ [])  = []
prodNil {suc (suc n)}  (A ∷ As)  = ( [] , prodNil As )

prodPair  : {n m : ℕ} → (As : Vec₁ Set n)
          → prodTy As 
          → prodTy (repeat₁ (Array m) ⊛₁ As) 
          → prodTy (repeat₁ (Array (suc m)) ⊛₁ As)
prodPair {0}            []        x1       v        = tt
prodPair {1}            (A ∷ [])  x1       v        = x1 ∷ v
prodPair {suc (suc n)}  (A ∷ As) (x , xs)  (v , vs) =
     x ∷ v , prodPair As xs vs

-- turn an array of products into a product of arrays by 
-- folding prodNil and prodPair on the vector.
transposeP   : {m : ℕ}{n : ℕ} → (As : Vec₁ Set n)
            → Array m (prodTy As) 
            → prodTy (repeat₁ (Array m) ⊛₁ As)
transposeP {zero}  As []        = prodNil As
transposeP {suc m} As (x ∷ xs)  = prodPair As  x (transposeP As xs)

defArrayAux : {m : ℕ} (n : ℕ) 
  → (As : Vec₁ Set (suc n)) → (NGsplit As) 
  → NGsplit (repeat₁ (Array m) ⊛₁ As)
defArrayAux n (A ∷ As) fa = \ x → transposeP As (map fa x)
   
-- uncurrying above definition
defArray  : (n : ℕ) → (m : ℕ)
          → NGsplit ⟨ ⋆ ⇒ ⋆ ⟩ (repeat₁ {suc n} (Array m))
defArray n m = k-curry (⋆ ⇒ ⋆) (defArrayAux {m} n)

